@testset "Test 2D" begin

    p = [0. 0. 1. 1. 0.5;  0. 1. 0. 1. 0.5] # nodes
    u = [0.0; 0.0; 0.0; 0.0; 1.0]   # function values in nodes
    u2 = [0.0; 0.0; 0.0; 0.0; 2.0]  # second function values in nodes
    t = [0.5 0.5 0.499999; 0.5 0.499999 0.5]  # evaluation points

    dp = [0.5 0.5; 0.5 0.5]
    es = [1.0 0.0; 0.0 1.0]
    du = [0.; 100000.0]

    @testset "Test 2D-RK_H0 kernel" begin
        rk = RK_H0(0.001)
        s = interpolate(p, u, rk)
        σ = evaluate(s, t)
        @test σ[1] ≈ u[5]
        @test isapprox(σ[2], u[5], atol = 1e-5)
        @test isapprox(σ[3], u[5], atol = 1e-5)
        @test isapprox(σ[2], σ[3], atol = 1e-5)

        σ1 = evaluate(s, t[:,1])
        @test σ1[1] ≈ u[5]

        rk = RK_H0()
        s = prepare(p, rk) # prepare spline
        s = construct(s, u) # construct spline
        σ1 = evaluate(s, t) # evaluate spline in points
        @test σ[1] ≈ u[5]
        @test isapprox(σ[2], u[5], atol = 1e-5)
        @test isapprox(σ[3], u[5], atol = 1e-5)
        @test isapprox(σ[2], σ[3], atol = 1e-5)

        s = construct(s, u2)
        σ2 = evaluate(s, t)
        @test σ2[1] ≈ u2[5]
        @test isapprox(σ2[2], u2[5], atol = 1e-5)
        @test isapprox(σ2[3], u2[5], atol = 1e-5)
        @test isapprox(σ2[2], σ2[3], atol = 1e-5)

        cond = get_cond(s)
        @test cond ≈ 10.0

        eps = get_epsilon(s)
        @test isapprox(eps, 1.32, atol = 1e-2)

        est_eps = estimate_epsilon(p) # get estimation of the problem's Gram matrix condition number
        @test isapprox(est_eps, 1.32, atol = 1e-2)
    end

    @testset "Test 2D-RK_H1 kernel" begin
        rk = RK_H1(0.001)
        s = interpolate(p, u, rk)
        cond = get_cond(s)
        @test cond ≈ 1.e11

        σ = evaluate(s, t)
        @test isapprox(σ[1], u[5], atol = 1e-5)
        @test isapprox(σ[2], u[5], atol = 1e-5)
        @test isapprox(σ[3], u[5], atol = 1e-5)
        @test isapprox(σ[2], σ[3], atol = 1e-5)

        grad = evaluate_grad(s, p[:,5])
        @test abs(grad[1]) < 1.0e-5 && abs(grad[2]) < 1.0e-5

        rk = RK_H1()
        s = prepare(p, rk) # prepare spline
        s = construct(s, u) # construct spline

        cond = get_cond(s)
        @test cond ≈ 100.0

        σ1 = evaluate(s, t) # evaluate spline in points
        @test isapprox(σ[1], u[5], atol = 1e-5)
        @test isapprox(σ[2], u[5], atol = 1e-5)
        @test isapprox(σ[3], u[5], atol = 1e-5)
        @test isapprox(σ[2], σ[3], atol = 1e-5)

        s = construct(s, u2)
        σ2 = evaluate(s, t)
        @test isapprox(σ[1], u[5], atol = 1e-5)
        @test isapprox(σ2[2], u2[5], atol = 1e-5)
        @test isapprox(σ2[3], u2[5], atol = 1e-5)
        @test isapprox(σ2[2], σ2[3], atol = 1e-5)

        est_eps = estimate_epsilon(p, dp) # get estimation of the problem's Gram matrix condition number
        @test est_eps ≈ 1.37 atol = 1e-2
###
        eps = 0.0001
        rk = RK_H1(eps)
        s = interpolate(p, u, dp, es, du, rk)
        cond = get_cond(s)
        @test cond == 1.0e14

        σ = evaluate(s, p[:,5])
        @test !isapprox(σ[1], u[5], atol = 0.1)

        q = estimate_interpolation_quality(s)
        @test q > 1.0

# Same test with extended precision
        rk = RK_H1(Double64(eps))
        p = Double64.(p)
        dp = Double64.(dp)
        es = Double64.(es)
        u = Double64.(u)
        du = Double64.(du)
        t = Double64.(t)
        s = prepare(p, dp, es, rk)
        s = construct(s, u, du)
        σ = evaluate(s, p)
        @test all(isapprox.(σ, u, atol = 1e-5))

        q = estimate_interpolation_quality(s)
        @test q < 1e-5
    end

    @testset "Test 2D-RK_H2 kernel" begin
        p = [0. 0. 1. 1. 0.5;  0. 1. 0. 1. 0.5] # nodes
        u = [0.0; 0.0; 0.0; 0.0; 1.0]   # function values in nodes
        u2 = [0.0; 0.0; 0.0; 0.0; 2.0]  # second function values in nodes
        t = [0.5 0.5 0.499999; 0.5 0.499999 0.5]  # evaluation points

        dp = [0.5 0.5; 0.5 0.5]
        es = [1.0 0.0; 0.0 1.0]
        du = [0.; 100000.0]

        rk = RK_H2(0.01)
        s = interpolate(p, u, rk)
        cond = get_cond(s)
        @test cond ≈ 1e11

        σ = evaluate(s, t)
        @test isapprox(σ[1], u[5], atol = 1e-5)
        @test isapprox(σ[2], u[5], atol = 1e-5)
        @test isapprox(σ[3], u[5], atol = 1e-5)
        @test isapprox(σ[2], σ[3], atol = 1e-5)

        grad = evaluate_grad(s, p[:,5])
        @test abs(grad[1]) < 1.0e-5 && abs(grad[2]) < 1.0e-5

        rk = RK_H2()
        s = prepare(p, rk) # prepare spline
        s = construct(s, u) # construct spline

        cond = get_cond(s)
        @test cond ≈ 1000.0

        σ1 = evaluate(s, t) # evaluate spline in points
        @test isapprox(σ[1], u[5], atol = 1e-5)
        @test isapprox(σ[2], u[5], atol = 1e-5)
        @test isapprox(σ[3], u[5], atol = 1e-5)
        @test isapprox(σ[2], σ[3], atol = 1e-5)

        s = construct(s, u2)
        σ2 = evaluate(s, t)
        @test isapprox(σ[1], u[5], atol = 1e-5)
        @test isapprox(σ2[2], u2[5], atol = 1e-5)
        @test isapprox(σ2[3], u2[5], atol = 1e-5)
        @test isapprox(σ2[2], σ2[3], atol = 1e-5)
###
        eps = 0.01
        rk = RK_H2(eps)
        s = interpolate(p, u, dp, es, du, rk)
        cond = get_cond(s)
        @test cond == 1.0e13

        σ = evaluate(s, p[:,5])
        @test !isapprox(σ[1], u[5], atol = 0.1)

        q = estimate_interpolation_quality(s)
        @test q > 1.0

# Same tests with extended precision
        rk = RK_H2(Double64(eps))
        p = Double64.(p)
        dp = Double64.(dp)
        es = Double64.(es)
        u = Double64.(u)
        du = Double64.(du)
        t = Double64.(t)
        s = prepare(p, dp, es, rk)
        s = construct(s, u, du)

        σ = evaluate(s, p)
        @test all(isapprox.(σ, u, atol = 1e-5))

        q = estimate_interpolation_quality(s)
        @test q < 1e-5
    end
end
